
/*******************************************************************************
*              (c), Copyright 2001, Marvell International Ltd.                 *
* THIS CODE CONTAINS CONFIDENTIAL INFORMATION OF MARVELL SEMICONDUCTOR, INC.   *
* NO RIGHTS ARE GRANTED HEREIN UNDER ANY PATENT, MASK WORK RIGHT OR COPYRIGHT  *
* OF MARVELL OR ANY THIRD PARTY. MARVELL RESERVES THE RIGHT AT ITS SOLE        *
* DISCRETION TO REQUEST THAT THIS CODE BE IMMEDIATELY RETURNED TO MARVELL.     *
* THIS CODE IS PROVIDED "AS IS". MARVELL MAKES NO WARRANTIES, EXPRESS, IMPLIED *
* OR OTHERWISE, REGARDING ITS ACCURACY, COMPLETENESS OR PERFORMANCE.           *
********************************************************************************
*******************************************************************************/



/******************************************************************************
* Introduction
* This source code is an implementation of UART over the MPSC port located on 
* the EV-64260A-B-0. This code helps implement customized code to use
* the UART over the MPSC port instead of the external UART and is implemented 
* without calling any operating system (OS) functions. This allows for easy
* porting to any operating system.
*
* The source code works in a polling mode.
*
*
* UART Over MPSC Source Code Description
*
* Files
*   uMpsc.h: Header file containing all macros and functions.
*   uMpsc.c: Source file with functions implementation.
*   core.h: Basic core functionality , includes read and write to 
*   	internal registers.
*  gt64260Ar.h: Internal registers offsets macros.
*
* System Environment
* This source code is not OS dependent because the memory allocation is
* completed without calling OS functions.
*
* Before using the code, change the defined value in the BASE_ADDRESS_RX_DESCRIPTORS 
* macro, located in uMpsc.h, to any SDRAM memory location that is free and uncached.
*
* Using the Functions
*   uartOverMpscInit: Only called once, when initializing the device.
*   uartOverMpscGetc: Returns the first character in the Rx queue. 
*   		  If the Rx queue is empty, this function returns 0xff.
*   uartOverMpscGetChar: Returns the first character in the Rx queue.
*   		     If Rx queue is empty, the function waits until a 
*   		     character is received in the Rx queue.
*   uartOverMpscPutc: Sends the character out to the Tx queue.
*
*/

#include "uMpsc.h"


/* this array hold the transmit descriptors chain */                  
static TX_DESC* txDescriptors[NUMBER_OF_TX_DESCRIPTORS];
/* hold the current processing transmit descriptor */
int txQueueIndex; 
                    

/* this array hold the receive descriptors chain */                  
static RX_DESC* rxDescriptors[NUMBER_OF_RX_DESCRIPTORS];
/* hold the current processing receive descriptor */
int rxQueueIndex;

/* the MPSC port number (0 or 1) */
int uartPort;
                                 

/******************************************************************************
* uartOverMpscInit - initialize the MPSC port to UART with baud rate of 115200.
*
* Input : port - one of the 2 MPSC ports (0-1).
*
* Output : The MPSC port is initialized to UART with baud rate of 115200.   
*
* RETURN true if success , false if can't get to hunt state.                                                                      
*/
bool uartOverMpscInit(int port)
{
unsigned int tempRegValue;
int i;
unsigned int confMpsc;            
unsigned int l_MpscCommand;
int timeOut = 100;

	txQueueIndex = 0;
	rxQueueIndex = 0;                     
	uartPort = port;

    /* initialize the communication unit */
	GT_REG_WRITE(MAIN_ROUTING_REGISTER,MRR_REG_VALUE);
    GT_REG_WRITE(SERIAL_PORT_MULTIPLEX,SERIAL_PORT_MULTIPLEX_REGISTER_VALUE); 

	/* config RCRR to take the BRG from BRG#0 */
	tempRegValue = 0;
    GT_REG_WRITE(RECEIVE_CLOCK_ROUTING_REGISTER,tempRegValue);    
    GT_REG_WRITE(TRANSMIT_CLOCK_ROUTING_REGISTER,tempRegValue); 

	/* allocate rx descriptors */                             
	for(i = 0 ; i < NUMBER_OF_RX_DESCRIPTORS ; i++)
	{
		rxDescriptors[i] = (RX_DESC*)(BASE_ADDRESS_RX_DESCRIPTORS + i*RX_DESCRIPTOR_SIZE);
        /* make the buffer size 1 */
		rxDescriptors[i]->bufsize = 1;
        /* set owner bit to the GT */
		rxDescriptors[i]->cmd_sts = OWNER_BY_GT | ENABLE_INTERRUPT | LAST | FIRST;
        /* put the buffer after the descriptor */
		rxDescriptors[i]->buffer = (BASE_ADDRESS_RX_DESCRIPTORS + i*RX_DESCRIPTOR_SIZE + 0x10);
		rxDescriptors[i]->buf_ptr = VIRTUAL_TO_PHY(rxDescriptors[i]->buffer);
        /* update the pointer to next descriptor */
		rxDescriptors[i]->next_desc_ptr = VIRTUAL_TO_PHY((BASE_ADDRESS_RX_DESCRIPTORS + (i+1)*RX_DESCRIPTOR_SIZE));

	}
    /* make it a cyclic descriptors chain */
	rxDescriptors[NUMBER_OF_RX_DESCRIPTORS-1]->next_desc_ptr = (unsigned int)VIRTUAL_TO_PHY(BASE_ADDRESS_RX_DESCRIPTORS);

	/* allocate tx descriptors */
	for(i = 0 ; i < NUMBER_OF_TX_DESCRIPTORS ; i++)
	{
		txDescriptors[i] = (TX_DESC*)(BASE_ADDRESS_TX_DESCRIPTORS + i*TX_DESCRIPTOR_SIZE);
        txDescriptors[i]->cmd_sts = 0; /* clean up */
        /* put the buffer after the descriptor */
		txDescriptors[i]->buffer = (BASE_ADDRESS_TX_DESCRIPTORS + i*TX_DESCRIPTOR_SIZE + 0x10);
		txDescriptors[i]->buf_ptr = VIRTUAL_TO_PHY(txDescriptors[i]->buffer);
        /* update the pointer to next descriptor */
		txDescriptors[i]->next_desc_ptr = VIRTUAL_TO_PHY((BASE_ADDRESS_TX_DESCRIPTORS + (i+1)*TX_DESCRIPTOR_SIZE));

	}
    /* make it a cyclic descriptors chain */
	txDescriptors[NUMBER_OF_TX_DESCRIPTORS-1]->next_desc_ptr = (unsigned int)VIRTUAL_TO_PHY(txDescriptors[0]);
    
    /* write the pointer to CDP */
    GT_REG_WRITE(CHANNEL0_CURRENT_RX_DESCRIPTOR_POINTER + CHANNELS_DIFFERENCE_OFFSETS*port,(unsigned int)VIRTUAL_TO_PHY(rxDescriptors[0]));
    
    /* write the tx desc pointer CDP and FDP */
    GT_REG_WRITE(CHANNEL0_FIRST_TX_DESCRIPTOR_POINTER + CHANNELS_DIFFERENCE_OFFSETS*port,(unsigned int)VIRTUAL_TO_PHY(txDescriptors[0]));
    GT_REG_WRITE(CHANNEL0_CURRENT_TX_DESCRIPTOR_POINTER + CHANNELS_DIFFERENCE_OFFSETS*port,(unsigned int)VIRTUAL_TO_PHY(txDescriptors[0]));

    /* changing to single frame mode confMpsc */
    GT_REG_READ(CHANNEL0_CONFIGURATION_REGISTER + CHANNELS_DIFFERENCE_OFFSETS*port,&confMpsc);
    confMpsc = confMpsc | SINGLE_FRAME_MODE | RECEIVE_FIFO_THRESHOLD | BURST_SIZE;
    GT_REG_WRITE(CHANNEL0_CONFIGURATION_REGISTER + CHANNELS_DIFFERENCE_OFFSETS*port,confMpsc);

    /* enable receive DMA */
    l_MpscCommand = ENABLE_RX_DMA;
    GT_REG_WRITE(CHANNEL0_COMMAND_REGISTER + CHANNELS_DIFFERENCE_OFFSETS*port,l_MpscCommand);

	/* set BRG#0 to 115200 from BCLK_in */
	GT_REG_WRITE (BRG0_CONFIGURATION_REGISTER,DEFAULT_UART_BRG);

	/* initialize the MPSC machine */
    GT_REG_WRITE(MPSC0_MAIN_CONFIGURATION_LOW+MPSC_GAP*port,DEFAULT_UART_MMCRL);
    GT_REG_WRITE(MPSC0_MAIN_CONFIGURATION_HIGH+MPSC_GAP*port,DEFAULT_UART_MMCRH);

    /* write to CHR1 */
    GT_REG_WRITE(CHANNEL0_REGISTER1+MPSC_GAP*port,DEFAULT_UART_CHR1);

    GT_REG_WRITE(CHANNEL0_REGISTER3+MPSC_GAP*port,DEFAULT_UART_CHR3); /* write to CHR3 */
    GT_REG_WRITE(CHANNEL0_REGISTER4+MPSC_GAP*port,DEFAULT_UART_CHR4); /* write to CHR4 */
    GT_REG_WRITE(CHANNEL0_REGISTER6+MPSC_GAP*port,DEFAULT_UART_CHR6); /* write to CHR6 */
    GT_REG_WRITE(CHANNEL0_REGISTER7+MPSC_GAP*port,DEFAULT_UART_CHR7); /* write to CHR7 */

    GT_REG_WRITE(CHANNEL0_REGISTER2+MPSC_GAP*port,DEFAULT_UART_CHR2); /* write to CHR2 */

    /* set the protocol configuration register */
    GT_REG_WRITE(MPSC0_PROTOCOL_CONFIGURATION+MPSC_GAP*port,DEFAULT_UART_MPCR);


	return true;
}

/******************************************************************************
* fixAddress - receive a buffer address and return the location of the first 
* char in the address - because Endians.
*/
unsigned int fixAddressOf64Bit(unsigned int addr)
{
    unsigned int lasrChar;

    lasrChar = addr & 0xf;
    if(((lasrChar >= 0)&&(lasrChar <= 3)) || 
       ((lasrChar >= 8)&&(lasrChar <= 0xb)))
        return addr+4;
    else
        return addr-4;

}




/******************************************************************************
* raiseTxCommandStatus - write to the tx command status.
*
* Input : address - pointer to address.
*
* Output : The owner bit is set to be owned by the GT.   
*
*/
void raiseTxCommandStatus(unsigned int* address)
{
    *address = LAST | FIRST | ENABLE_INTERRUPT | OWNER_BY_GT | PREAMBLE;
}



/******************************************************************************
* uartOverMpscGetc - get one character from the MPSC port.
*
* Output : return the first char in the queue.   
* if empty - return 0xff.
*/
char uartOverMpscGetc()
{
RX_DESC* pCurrDesc;
unsigned int tempBuffer;


	pCurrDesc = rxDescriptors[rxQueueIndex];
    if(pCurrDesc->cmd_sts & OWNER_BY_GT)
        {
            return (char)0xff;
        }

    tempBuffer = *(unsigned int*)PHY_TO_VIRTUAL(fixAddressOf64Bit(pCurrDesc->buf_ptr));

	*(unsigned int*)(&(pCurrDesc->cmd_sts)) = OWNER_BY_GT | ENABLE_INTERRUPT | LAST | FIRST;
	rxQueueIndex = (rxQueueIndex+1) % NUMBER_OF_RX_DESCRIPTORS; /* point to next descriptor */

    return (char)(tempBuffer & 0xff);


}



/******************************************************************************
* uartOverMpscGetChar - get one character from the MPSC port.
*
* Output : return the first char in the queue.   
* if empty - wait for char.
*/
char uartOverMpscGetChar()
{
RX_DESC* pCurrDesc;
unsigned int tempBuffer;


	pCurrDesc = rxDescriptors[rxQueueIndex];
    while(pCurrDesc->cmd_sts & OWNER_BY_GT);

    tempBuffer = *(unsigned int*)PHY_TO_VIRTUAL(fixAddressOf64Bit(pCurrDesc->buf_ptr));

	*(unsigned int*)(&(pCurrDesc->cmd_sts)) = OWNER_BY_GT | ENABLE_INTERRUPT | LAST | FIRST;
	rxQueueIndex = (rxQueueIndex+1) % NUMBER_OF_RX_DESCRIPTORS; /* point to next descriptor */

    return (char)(tempBuffer & 0xff);


}


/******************************************************************************
* uartOverMpscPutc - put out a character on the MPSC port.
*
* Output : one character is transmitted out.   
*
* RETURN true if success , false if the tx queue is full.                                                                      
*/
bool uartOverMpscPutc(char charToPut)
{
TX_DESC* pCurrTxDesc;
unsigned int txCmd;

	pCurrTxDesc = txDescriptors[txQueueIndex];
    while(pCurrTxDesc->cmd_sts & OWNER_BY_GT);

    *(unsigned int*)PHY_TO_VIRTUAL(fixAddressOf64Bit(pCurrTxDesc->buf_ptr)) = charToPut;
    pCurrTxDesc->bytecnt = 1;

    raiseTxCommandStatus((unsigned int*)(&(pCurrTxDesc->cmd_sts)));

	GT_REG_READ(CHANNEL0_COMMAND_REGISTER + CHANNELS_DIFFERENCE_OFFSETS*uartPort,&txCmd);
    if(!(txCmd & TX_DEMAND) )
    {
        txCmd = txCmd | TX_DEMAND;
        GT_REG_WRITE(CHANNEL0_COMMAND_REGISTER + CHANNELS_DIFFERENCE_OFFSETS*uartPort,txCmd);
    }
			 
	/* update the current descriptor */
	txQueueIndex = (txQueueIndex+1) % NUMBER_OF_TX_DESCRIPTORS; /* point to next */

	return true;
}
